import typing

from sb_json_tools import jsondiff
import logging

from karp import lex
from karp.lex import GetHistoryDto, HistoryDto

from karp.lex.domain.entities import Entry
from karp.lex.application.queries import EntryViews, EntryDto, EntryDiffDto, GetEntryDiff, GetEntryHistory, GetHistory, EntryHistoryRequest, EntryDiffRequest, GetEntryRepositoryId
from karp.foundation.value_objects import unique_id
from karp.lex.application.repositories import EntryUowRepositoryUnitOfWork


logger = logging.getLogger(__name__)


class GenericEntryViews(EntryViews):
    def __init__(
        self,
        get_entry_repo_id: GetEntryRepositoryId,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> None:
        super().__init__()
        self.get_entry_repo_id = get_entry_repo_id
        self.entry_repo_uow = entry_repo_uow

    def get_by_id(self, resource_id: str, entity_id: unique_id.UniqueId) -> EntryDto:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return self._entry_to_entry_dto(
                uw.repo.by_id(entity_id),
                resource_id
            )

    def get_by_id_optional(
        self,
        resource_id: str,
        entity_id: unique_id.UniqueId
    ) -> typing.Optional[EntryDto]:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            entry = uw.repo.get_by_id_optional(entity_id)
            if entry:
                return self._entry_to_entry_dto(entry, resource_id)
        return None

    def get_by_entry_id(self, resource_id: str, entry_id: str) -> EntryDto:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return self._entry_to_entry_dto(
                uw.repo.by_entry_id(entry_id),
                resource_id
            )

    def get_by_entry_id_optional(
        self,
        resource_id: str,
        entry_id: str
    ) -> typing.Optional[EntryDto]:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            entry = uw.repo.get_by_entry_id_optional(entry_id)
            if entry:
                return self._entry_to_entry_dto(entry, resource_id)
        return None

    def get_total(self, resource_id: str) -> int:
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return uw.repo.get_total_entries()

    def get_by_referenceable(self, resource_id: str, filters):
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return (
                self._entry_to_entry_dto(entry, resource_id)
                for entry in uw.repo.by_referenceable(filters)
            )

    def all_entries(self, resource_id: str):
        entry_repo_id = self.get_entry_repo_id.query(resource_id)
        with self.entry_repo_uow as uw:
            entry_uow = uw.repo.get_by_id(entry_repo_id)
        with entry_uow as uw:
            return (
                self._entry_to_entry_dto(entry, resource_id)
                for entry in uw.repo.all_entries()
            )

    def _entry_to_entry_dto(self, entry: Entry, resource_id: str) -> EntryDto:
        return EntryDto(
            entry_id=entry.entry_id,
            entity_id=entry.entity_id,
            resource=resource_id,
            version=entry.version,
            entry=entry.body,
            last_modified=entry.last_modified,
            last_modified_by=entry.last_modified_by,
        )


class GenericEntryQuery:
    def __init__(
        self,
        resource_uow: lex.ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ):
        self._resource_uow = resource_uow
        self.entry_repo_uow = entry_repo_uow

    def get_entry_repo_id(self, resource_id: str) -> unique_id.UniqueId:
        with self._resource_uow as uw:
            return uw.repo.by_resource_id(resource_id).entry_repository_id


class GenericGetEntryHistory(GenericEntryQuery, GetEntryHistory):
    def query(
        self,
        resource_id: str,
        entry_id: str,
        version: typing.Optional[int],
    ) -> EntryDto:
        entry_repo_id = self.get_entry_repo_id(resource_id)
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(entry_repo_id) as uw:
            result = uw.repo.by_entry_id(entry_id, version=version)

        return EntryDto(
            entry_id=entry_id,
            entity_id=result.id,
            resource=resource_id,
            version=result.version,
            entry=result.body,
            last_modified_by=result.last_modified_by,
            last_modified=result.last_modified,
        )


class GenericGetHistory(GenericEntryQuery, GetHistory):
    def query(
        self,
        history_request: EntryHistoryRequest,
    ) -> GetHistoryDto:
        logger.info('querying history', extra={
                    'history_request': history_request})
        entry_repo_id = self.get_entry_repo_id(history_request.resource_id)
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(entry_repo_id) as uw:
            paged_query, total = uw.repo.get_history(
                entry_id=history_request.entry_id,
                user_id=history_request.user_id,
                from_date=history_request.from_date,
                to_date=history_request.to_date,
                from_version=history_request.from_version,
                to_version=history_request.to_version,
                offset=history_request.current_page * history_request.page_size,
                limit=history_request.page_size,
            )
        result = []
        previous_body = {}
        for history_entry in paged_query:
            # TODO fix this, we should get the diff in another way, probably store the diffs directly in the database
            # entry_version = history_entry.version
            # if entry_version > 1:
            #     previous_entry = uw.repo.by_entry_id(
            #         history_entry.entry_id, version=entry_version - 1
            #     )
            #     previous_body = previous_entry.body
            # else:
            #     previous_body = {}
            history_diff = jsondiff.compare(previous_body, history_entry.body)
            logger.info('diff', extra={'diff': history_diff})
            result.append(
                HistoryDto(
                    timestamp=history_entry.last_modified,
                    message=history_entry.message or "",
                    entry_id=history_entry.entry_id,
                    version=history_entry.version,
                    op=history_entry.op,
                    user_id=history_entry.last_modified_by,
                    diff=history_diff,
                )
            )
            previous_body = history_entry.body

        return GetHistoryDto(history=result, total=total)


class GenericGetEntryDiff(GenericEntryQuery, GetEntryDiff):
    def query(
        self,
        diff_request: EntryDiffRequest,
    ) -> EntryDiffDto:

        entry_repo_id = self.get_entry_repo_id(diff_request.resource_id)
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(entry_repo_id) as uw:
            db_entry = uw.repo.by_entry_id(diff_request.entry_id)

            #     src = resource_obj.model.query.filter_by(entry_id=entry_id).first()
            #
            #     query = resource_obj.history_model.query.filter_by(entry_id=src.id)
            #     timestamp_field = resource_obj.history_model.timestamp
            #
            if diff_request.from_version:
                obj1 = uw.repo.by_id(
                    db_entry.id, version=diff_request.from_version)
            elif diff_request.from_date is not None:
                obj1 = uw.repo.by_id(
                    db_entry.id, after_date=diff_request.from_date)
            else:
                obj1 = uw.repo.by_id(db_entry.id, oldest_first=True)

            obj1_body = obj1.body if obj1 else None

            if diff_request.to_version:
                obj2 = uw.repo.by_id(
                    db_entry.id, version=diff_request.to_version)
                obj2_body = obj2.body
            elif diff_request.to_date is not None:
                obj2 = uw.repo.by_id(
                    db_entry.id, before_date=diff_request.to_date)
                obj2_body = obj2.body
            elif diff_request.entry is not None:
                obj2 = None
                obj2_body = diff_request.entry
            else:
                obj2 = db_entry
                obj2_body = db_entry.body

        #     elif from_date is not None:
        #         obj1_query = query.filter(timestamp_field >= from_date).order_by(
        #             timestamp_field
        #         )
        #     else:
        #         obj1_query = query.order_by(timestamp_field)
        #     if to_version:
        #         obj2_query = query.filter_by(version=to_version)
        #     elif to_date is not None:
        #         obj2_query = query.filter(timestamp_field <= to_date).order_by(
        #             timestamp_field.desc()
        #         )
        #     else:
        #         obj2_query = None
        #
        #     obj1 = obj1_query.first()
        #     obj1_body = json.loads(obj1.body) if obj1 else None
        #
        #     if obj2_query:
        #         obj2 = obj2_query.first()
        #         obj2_body = json.loads(obj2.body) if obj2 else None
        #     elif entry is not None:
        #         obj2 = None
        #         obj2_body = entry
        #     else:
        #         obj2 = query.order_by(timestamp_field.desc()).first()
        #         obj2_body = json.loads(obj2.body) if obj2 else None
        #
        if not obj1_body or not obj2_body:
            raise karp_errors.KarpError("diff impossible!")
        #
        return EntryDiffDto(
            diff=jsondiff.compare(obj1_body, obj2_body),
            from_version=obj1.version,
            to_version=obj2.version if obj2 else None,
        )



import pkg_resources

# from flask import Flask  # pyre-ignore
# from flask_cors import CORS  # pyre-ignore
# from flask import request  # pyre-ignore
# import flask_reverse_proxy
# import werkzeug.exceptions


__version__ = "6.0.7"


# TODO handle settings correctly
# def create_app(config_class=None):
#     app = Flask(__name__)
#     app.config.from_object("karp.config.DevelopmentConfig")
#     if config_class:
#         app.config.from_object(config_class)
#     if os.getenv("KARP_CONFIG"):
#         app.config.from_object(os.getenv("KARP_CONFIG"))

#     logger = setup_logging(app)

#     from .api import (
#         health_api,
#         edit_api,
#         query_api,
#         conf_api,
#         documentation,
#         stats_api,
#         history_api,
#     )

#     app.register_blueprint(edit_api)
#     app.register_blueprint(health_api)
#     app.register_blueprint(query_api)
#     app.register_blueprint(conf_api)
#     app.register_blueprint(documentation)
#     app.register_blueprint(stats_api)
#     app.register_blueprint(history_api)

#     from .database import db

#     db.init_app(app)

#     if app.config.get("SETUP_DATABASE", True):
#         from .resourcemgr import setup_resource_classes

#         with app.app_context():
#             setup_resource_classes()

#     if app.config["ELASTICSEARCH_ENABLED"] and app.config.get("ELASTICSEARCH_HOST", ""):
#         from karp.elasticsearch import init_es

#         init_es(app.config["ELASTICSEARCH_HOST"])
#     else:
#         # TODO if an elasticsearch test runs before a non elasticsearch test this
#         # is needed to reset the index and search modules
#         from karp import search
#         from karp.indexmgr.index import IndexInterface
#         from karp.indexmgr import indexer

#         search.init(search.SearchInterface())
#         indexer.init(IndexInterface())

#     with app.app_context():
#         import karp.pluginmanager

#         karp.pluginmanager.init()

#     @app.errorhandler(Exception)
#     def http_error_handler(error: Exception):
#         error_str = "Exception on %s [%s]" % (request.path, request.method)
#         if isinstance(error, werkzeug.exceptions.NotFound):
#             logger.debug(error_str)
#             return "", 404

#         if isinstance(error, KarpError):
#             logger.debug(error_str)
#             logger.debug(error.message)
#             error_code = error.code if error.code else 0
#             return (
#                 json.dumps({"error": error.message, "errorCode": error_code}),
#                 error.http_return_code,
#             )
#         else:
#             if app.config["DEBUG"]:
#                 raise error
#             logger.error(error_str)
#             logger.exception("unhandled exception")
#             return json.dumps({"error": "unknown error", "errorCode": 0}), 400

#     import karp.auth.auth as auth

#     if app.config["JWT_AUTH"]:
#         from karp.auth.jwt_authenticator import JWTAuthenticator

#         auth.auth.set_authenticator(JWTAuthenticator())
#     else:
#         from karp.auth.authenticator import Authenticator

#         auth.auth.set_authenticator(Authenticator())

#     CORS(app)

#     app.wsgi_app = flask_reverse_proxy.ReverseProxied(app.wsgi_app)
#     return app


# def setup_logging(app):
#     logger = logging.getLogger("karp")
#     if app.config.get("LOG_TO_SLACK"):
#         slack_handler = slack_logging.get_slack_logging_handler(
#             app.config.get("SLACK_SECRET")
#         )
#         logger.addHandler(slack_handler)
#     console_handler = logging.StreamHandler()
#     formatter = logging.Formatter(
#         "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
#     )
#     console_handler.setFormatter(formatter)
#     logger.setLevel(app.config["CONSOLE_LOG_LEVEL"])
#     logger.addHandler(console_handler)
#     return logger


def get_version() -> str:
    return __version__


def get_resource_string(name: str) -> str:
    return pkg_resources.resource_string(__name__, name).decode("utf-8")



"""Handles all sql db connections."""
import collections
import json
from typing import Any, Dict, Optional

import attr
import sqlalchemy
from sqlalchemy import (JSON, Column, Enum, ForeignKey, Integer, MetaData,
                        String, Table, Text, Unicode, and_, event, exc, func,
                        or_)
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm import aliased, mapper, relationship
from sqlalchemy.orm.session import Session, sessionmaker
from sqlalchemy.schema import (ForeignKeyConstraint, PrimaryKeyConstraint,
                               UniqueConstraint)
from sqlalchemy.sql import delete, insert, update
from sqlalchemy.types import VARCHAR, Boolean, Float, Time, TypeDecorator
from sqlalchemy_json import NestedMutableJson
from sqlalchemy_utils import UUIDType


# engine = sqlalchemy.create_engine(config.DB_URL, echo=True)

metadata = MetaData()

Base = declarative_base(metadata=metadata)

# SessionLocal = sessionmaker(bind=engine)


@attr.s(auto_attribs=True)
class SQLEngineSessionfactory:
    engine: Engine = attr.ib()
    session_factory: sessionmaker = attr.ib()
    metadata: MetaData = attr.ib()

    @classmethod
    def from_db_uri(cls, db_uri: str):
        engine = sqlalchemy.create_engine(db_uri)
        session_factory = sessionmaker(bind=engine)
        metadata_ = MetaData()
        return cls(engine, session_factory, metadata_)


_db_handlers: Dict[str, SQLEngineSessionfactory] = dict()


def create_session(db_uri: str) -> Session:
    _assure_in_db_handlers(db_uri)
    return _db_handlers[db_uri].session_factory()


def get_engine(db_uri: str) -> Engine:
    _assure_in_db_handlers(db_uri)
    return _db_handlers[db_uri].engine


def _assure_in_db_handlers(db_uri: str):
    if db_uri not in _db_handlers:
        _db_handlers[db_uri] = SQLEngineSessionfactory.from_db_uri(db_uri)


def _metadata(db_uri: str) -> MetaData:
    _assure_in_db_handlers(db_uri)
    return _db_handlers[db_uri].metadata


def get_table(table_name: str) -> Optional[Table]:
    return metadata.tables.get(table_name)


# def map_class_to_some_table(cls, table: Table, entity_name: str, **mapper_kw):
#     """Use the EntityName-pattern to map a cls to several tables.

#     See https://github.com/sqlalchemy/sqlalchemy/wiki/EntityName for more information.

#     Arguments:
#         table {Table} -- the table to map to
#         entity_name {str} -- the name for the new class

#     Keyword arguments:
#     mapper_kw -- any keyword arguments passed to mapper

#     Raises:
#         ValueError: when?

#     Returns:
#         Any -- the new class constructed
#     """
#     newcls = type(entity_name, (cls,), {})
#     mapper(newcls, table, **mapper_kw)
#     return newcls


# _metadata = MetaData()


# class JsonEncodedDict(TypeDecorator):
#     """Represent an immutable dictionary as a json-encoded-string."""

#     impl = VARCHAR

#     def process_bind_param(self, value, dialect):
#         if value is not None:
#             value = json.dumps(value)
#         return value

#     def process_result_value(self, value, dialect):
#         if value is not None:
#             value = json.loads(value)
#         return value


# class MutableDict(Mutable, dict):
#     """Tracking changes for sql."""

#     @classmethod
#     def coerce(cls, key, value):
#         """Convert plain dicts to MutableDict."""
#         if not isinstance(value, MutableDict):
#             if isinstance(value, dict):
#                 return cls(value)

#             # this call will raise ValueError
#             return Mutable.coerce(key, value)
#         else:
#             return value

#     def __setitem__(self, key, value):
#         """Detect dictionary set events and emit changed event."""
#         dict.__setitem__(self, key, value)
#         self.changed()

#     def __delitem__(self, key, value):
#         """Detect dictionary del events and emit changed event."""
#         dict.__delitem__(self, key, value)
#         self.changed()


# MutableDict.associate_with(JsonEncodedDict)


# class FalseEncodedAsNullBoolean(TypeDecorator):
#     impl = Boolean

#     def process_bind_param(self, value, dialect):
#         if value is False:
#             value = None
#         return value

#     def process_result_value(self, value, dialect):
#         if value is None:
#             return False
#         return value


# class MutableBool(Mutable):
#     @classmethod
#     def coerce(cls, key, value):
#         if isinstance(value, MutableBool):
#             return value
#
#         if isinstance(value, bool):
#             return cls(value)
#
#         return Mutable.coerce(key, value)
#
#     def __setattr__(self, name, value):
#         object.__setattr__(self, name, value)
#         self.changed()


# MutableBool.associate_with(FalseEncodedAsNullBoolean)



import logging
from pathlib import Path
from typing import Optional

import json_streams
import typer
from tabulate import tabulate
from tqdm import tqdm

from karp.foundation.commands import CommandBus
from karp.lex.domain import commands
# from karp.lex.domain.errors import ResourceAlreadyPublished

from .utility import cli_error_handler, cli_timer
from .typer_injector import inject_from_ctx

logger = logging.getLogger("karp")


subapp = typer.Typer()


@subapp.command("import")
@cli_error_handler
@cli_timer
def import_resource(
    resource_id: str,
    # version: Optional[int],
    data: Path,
    ctx: typer.Context,
):
    bus = inject_from_ctx(CommandBus, ctx)
    cmd = commands.AddEntries(
        resource_id=resource_id,
        # entries=json_streams.load_from_file(data),
        entries=tqdm(
            json_streams.load_from_file(data),
            desc="Importing",
            unit=" entries"
        ),
        user="local admin",
        message="imported through cli",
    )
    bus.dispatch(cmd)
    typer.echo(f"Successfully imported entries to {resource_id}")


@subapp.command("update")
@cli_error_handler
@cli_timer
def update_entries(resource_id: str, data: Path):
    updated_entries = entries.update_entries(
        resource_id, json_streams.load_from_file(data)
    )


def init_app(app):
    app.add_typer(subapp, name="entries")



import json
import logging
import typing
from pathlib import Path
from typing import IO, Dict, Generic, List, Optional, Tuple

import logging

from karp import errors as karp_errors
from karp.lex.domain import errors, events, entities
from karp.lex.domain.entities import Resource
from karp.foundation import events as foundation_events
from karp.foundation.commands import CommandHandler
from karp.foundation import messagebus
from karp.lex.application.queries import ResourceDto
from karp.lex.application import repositories as lex_repositories
from karp.lex.domain import commands
from karp.lex.application import repositories


logger = logging.getLogger(__name__)

resource_models = {}  # Dict
history_models = {}  # Dict
resource_configs = {}  # Dict
resource_versions = {}  # Dict[str, int]
field_translations = {}  # Dict[str, Dict[str, List[str]]]


def get_field_translations(resource_id: str) -> Optional[Dict]:
    with repositories(using=ctx.resource_repo) as uw:
        resource = uw.by_resource_id(resource_id)
        return resource.config.get("field_mapping") if resource else None


class BasingResource:
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork
    ) -> None:
        self.resource_uow = resource_uow

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()

# def get_resource(resource_id: str, version: Optional[int] = None) -> Resource:
#     if not version:
#         resource_def = database.get_active_resource_definition(resource_id)
#         if not resource_def:
#             raise ResourceNotFoundError(resource_id)
#         if resource_id not in resource_models:
#             setup_resource_class(resource_id)
#         return Resource(
#             model=resource_models[resource_id],
#             history_model=history_models[resource_id],
#             resource_def=resource_def,
#             version=resource_versions[resource_id],
#             config=resource_configs[resource_id],
#         )
#     else:
#         resource_def = database.get_resource_definition(resource_id, version)
#         if not resource_def:
#             raise ResourceNotFoundError(resource_id, version=version)
#         config = json.loads(resource_def.config_file)
#         cls = database.get_or_create_resource_model(config, version)
#         history_cls = database.get_or_create_history_model(resource_id, version)
#         return Resource(
#             model=cls,
#             history_model=history_cls,
#             resource_def=resource_def,
#             version=version,
#             config=config,
#         )


# def get_all_resources() -> List[database.ResourceDefinition]:
#     return database.ResourceDefinition.query.all()


def check_resource_published(resource_ids: List[str]) -> None:
    published_resources = [
        resource.resource_id for resource in get_published_resources()
    ]
    for resource_id in resource_ids:
        if resource_id not in published_resources:
            raise errors.KarpError(
                'Resource is not searchable: "{resource_id}"'.format(
                    resource_id=resource_id
                ),
                errors.ClientErrorCodes.RESOURCE_NOT_PUBLISHED,
            )


# def create_and_update_caches(id: str, version: int, config: Dict) -> None:
#     resource_models[id] = database.get_or_create_resource_model(config, version)
#     history_models[id] = database.get_or_create_history_model(id, version)
#     resource_versions[id] = version
#     resource_configs[id] = config
#     if "field_mapping" in config:
#         field_translations[id] = config["field_mapping"]


# def remove_from_caches(id: str) -> None:
#     del resource_models[id]
#     del history_models[id]
#     del resource_versions[id]
#     del resource_configs[id]
#     if id in field_translations:
#         del field_translations[id]


# def setup_resource_classes() -> None:
#     for resource in get_available_resources():
#         config = json.loads(resource.config_file)
#         create_and_update_caches(config["resource_id"], resource.version, config)


# def setup_resource_class(resource_id, version=None):
#     if version:
#         resource_def = database.get_resource_definition(resource_id, version)
#     else:
#         resource_def = database.get_active_resource_definition(resource_id)
#     if not resource_def:
#         raise ResourceNotFoundError(resource_id, version)
#     config = json.loads(resource_def.config_file)
#     create_and_update_caches(resource_id, resource_def.version, config)


# def _read_and_process_resources_from_dir(
#     config_dir: str, func
# ) -> List[Tuple[str, int]]:
#     files = [f for f in os.listdir(config_dir) if re.match(r".*\.json$", f)]
#     result = []
#     for filename in files:
#         print("Processing '{filename}' ...".format(filename=filename))
#         with open(os.path.join(config_dir, filename)) as fp:
#             result.append(func(fp, config_dir=config_dir))
#     return result


# def create_new_resource_from_dir(config_dir: str) -> List[Tuple[str, int]]:
#     return _read_and_process_resources_from_dir(config_dir, create_new_resource)


def create_new_resource_from_file(config_file: IO) -> Resource:
    return create_new_resource(config_file)


def create_resource_from_path(config: Path) -> List[Resource]:
    return []


# def update_resource_from_dir(config_dir: str) -> List[Tuple[str, int]]:
#     return _read_and_process_resources_from_dir(config_dir, update_resource)


# def update_resource_from_file(config_file: BinaryIO) -> Tuple[str, int]:
#     return update_resource(config_file)

class CreatingResource(CommandHandler[commands.CreateResource], BasingResource):
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork,
        entry_repo_uow: lex_repositories.EntryUowRepositoryUnitOfWork
    ) -> None:
        super().__init__(resource_uow=resource_uow)
        self.entry_repo_uow = entry_repo_uow

    def execute(self, cmd: commands.CreateResource) -> ResourceDto:
        with self.entry_repo_uow as uow:
            entry_repo_exists = uow.repo.get_by_id_optional(cmd.entry_repo_id)
            if not entry_repo_exists:
                raise errors.NoSuchEntryRepository(
                    f"Entry repository '{cmd.entry_repo_id}' not found")

        with self.resource_uow as uow:
            existing_resource = uow.repo.get_by_resource_id_optional(
                cmd.resource_id)
            if (
                existing_resource
                and not existing_resource.discarded
                and existing_resource.entity_id != cmd.entity_id
            ):
                raise errors.IntegrityError(
                    f"Resource with resource_id='{cmd.resource_id}' already exists."
                )

            resource = entities.create_resource(
                entity_id=cmd.entity_id,
                resource_id=cmd.resource_id,
                config=cmd.config,
                message=cmd.message,
                entry_repo_id=cmd.entry_repo_id,
                created_at=cmd.timestamp,
                created_by=cmd.user,
                name=cmd.name,
            )

            uow.repo.save(resource)
            uow.commit()
        return ResourceDto(**resource.dict())

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


def setup_existing_resources(evt):
    with ctx.resource_uow:
        for resource_id in ctx.resource_uow.repo.resource_ids():
            resource = ctx.resource_uow.repo.by_resource_id(resource_id)
            entry_repo_uow = ctx.entry_uow_factory.create(
                resource_id=resource_id,
                resource_config=resource.config,
                entry_repository_settings=resource.entry_repository_settings,
            )
            ctx.entry_uows.set_uow(resource_id, entry_repo_uow)


def create_new_resource(config_file: IO, config_dir=None) -> Resource:
    logger.debug("loading config with json ...")
    config = json.load(config_file)
    logger.debug("creating resource ...")
    resource = Resource.from_dict(config)

    logger.debug("resource created. setting up repositories")
    with repositories(using=ctx.resource_repo) as uw:
        logger.debuf("adding resource to ctx.resource_repo ...")
        uw.put(resource)

    logger.debug("done! returning ...")
    return resource


#     config = load_and_validate_config(config_file)

#     resource_id = config["resource_id"]

#     version = database.get_next_resource_version(resource_id)

#     entry_json_schema = create_entry_json_schema(config)

#     config = load_plugins_to_config(config, version, config_dir)

#     resource = {
#         "resource_id": resource_id,
#         "version": version,
#         "config_file": json.dumps(config),
#         "entry_json_schema": json.dumps(entry_json_schema),
#     }

#     new_resource = database.ResourceDefinition(**resource)
#     db.session.add(new_resource)
#     db.session.commit()

#     sqlalchemyclass = database.get_or_create_resource_model(config, version)
#     history_model = database.get_or_create_history_model(resource_id, version)
#     sqlalchemyclass.__table__.create(bind=db.engine)
#     for child_class in sqlalchemyclass.child_tables.values():
#         child_class.__table__.create(bind=db.engine)
#     history_model.__table__.create(bind=db.engine)

#     return resource["resource_id"], resource["version"]


# def load_and_validate_config(config_file):
#     config = json.load(config_file)
#     try:
#         schema = get_resource_string("schema/resourceconf.schema.json")
#         validate_conf = fastjsonschema.compile(json.loads(schema))
#         validate_conf(config)
#     except fastjsonschema.JsonSchemaException as e:
#         raise ResourceInvalidConfigError(config["resource_id"], config_file, e.message)
#     return config


# def load_plugins_to_config(config: Dict, version, config_dir) -> Dict:
#     resource_id = config["resource_id"]

#     if "plugins" in config:
#         for plugin_id in config["plugins"]:
#             import karp.pluginmanager as plugins

#             if plugin_id not in plugins.plugins:
#                 raise PluginNotFoundError(plugin_id, resource_id)
#             for (field_name, field_conf) in plugins.plugins[
#                 plugin_id
#             ].resource_creation(resource_id, version, config_dir):
#                 config["fields"][field_name] = field_conf

#     return config

class UpdatingResource(CommandHandler[commands.UpdateResource], BasingResource):
    def __init__(self, resource_uow: repositories.ResourceUnitOfWork) -> None:
        super().__init__(resource_uow=resource_uow)

    def execute(self, cmd: commands.UpdateResource):
        with self.resource_uow as uow:
            resource = uow.repo.by_resource_id(cmd.resource_id)
            found_changes = False
            if resource.name != cmd.name:
                resource.name = cmd.name
                found_changes = True
            if resource.config != cmd.config:
                resource.config = cmd.config
                found_changes = True
            if found_changes:
                resource.stamp(
                    user=cmd.user,
                    message=cmd.message,
                    timestamp=cmd.timestamp,
                )
                uow.repo.save(resource)
            uow.commit()

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


# def update_resource(config_file: BinaryIO, config_dir=None) -> Tuple[str, int]:
#     config = load_and_validate_config(config_file)

#     resource_id = config["resource_id"]
#     entry_json_schema = create_entry_json_schema(config)

#     resource_def = database.get_active_or_latest_resource_definition(resource_id)
#     if not resource_def:
#         raise RuntimeError(
#             "Could not find a resource_definition with id '{resource_id}".format(
#                 resource_id=resource_id
#             )
#         )
#     config = load_plugins_to_config(config, resource_def.version, config_dir)

#     config_diff = jsondiff.compare(json.loads(resource_def.config_file), config)
#     needs_reindex = False
#     not_allowed_change = False
#     not_allowed_changes = []
#     changes = []
#     for diff in config_diff:
#         print("config_diff = '{diff}'".format(diff=diff))
#         changes.append(diff)

#         if diff["type"] == "TYPECHANGE":
#             not_allowed_change = True
#             not_allowed_changes.append(diff)
#         elif diff["field"].endswith("required"):
#             continue
#         elif diff["field"] == "field_mapping":
#             continue
#         elif diff["type"] == "ADDED":
#             # TODO add rule for field_mappings
#             needs_reindex = True
#         elif diff["type"] == "REMOVED":
#             needs_reindex = True
#         elif diff["type"] == "CHANGE":
#             if diff["field"].endswith("type"):
#                 not_allowed_change = True
#                 not_allowed_changes.append(diff)
#             else:
#                 needs_reindex = True
#         else:
#             not_allowed_change = True
#             not_allowed_changes.append(diff)

#     # entry_json_schema_diff = jsondiff.compare(json.loads(resource_def.entry_json_schema), entry_json_schema)
#     # for diff in entry_json_schema_diff:
#     #     print("entry_json_schema_diff = '{diff}'".format(diff=diff))

#     if not config_diff:
#         return resource_id, None

#     if not_allowed_change:
#         raise ResourceConfigUpdateError(
#             """
#             You must 'create' a new version of '{resource_id}', current version '{version}'.
#             Changes = {not_allowed_changes}""".format(
#                 resource_id=resource_id,
#                 version=resource_def.version,
#                 not_allowed_changes=not_allowed_changes,
#             ),
#             resource_id,
#             config_file,
#         )
#     if needs_reindex:
#         print(
#             "You might need to reindex '{resource_id}' version '{version}'".format(
#                 resource_id=resource_id, version=resource_def.version
#             )
#         )
#     else:
#         print("Safe update.")
#     print("Changes:")
#     for diff in changes:
#         print(" - {diff}".format(diff=diff))
#     resource_def.config_file = json.dumps(config)
#     resource_def.entry_json_schema = json.dumps(entry_json_schema)
#     # db.session.update(resource_def)
#     db.session.commit()
#     if resource_def.active:
#         create_and_update_caches(resource_id, resource_def.version, config)

#     return resource_id, resource_def.version

class PublishingResource(CommandHandler[commands.PublishResource], BasingResource):
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork,
        **kwargs,
    ) -> None:
        super().__init__(resource_uow=resource_uow)

    def execute(self, cmd: commands.PublishResource):
        logger.info('publishing resource', extra={
                    'resource_id': cmd.resource_id})
        with self.resource_uow as uow:
            resource = uow.repo.by_resource_id(cmd.resource_id)
            if not resource:
                raise errors.ResourceNotFound(cmd.resource_id)
            resource.publish(user=cmd.user, message=cmd.message,
                             timestamp=cmd.timestamp)
            uow.repo.save(resource)
            uow.commit()

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


#     resource = database.get_resource_definition(resource_id, version)
#     old_active = database.get_active_resource_definition(resource_id)
#     if old_active:
#         old_active.active = False
#     resource.active = True
#     db.session.commit()

#     config = json.loads(resource.config_file)
#     # this stuff doesn't matter right now since we are not modifying the state of the actual app, only the CLI
#     create_and_update_caches(resource_id, resource.version, config)


# def unpublish_resource(resource_id):
#     resource = database.get_active_resource_definition(resource_id)
#     if resource:
#         resource.active = False
#         db.session.update(resource)
#         db.session.commit()
#     remove_from_caches(resource_id)


# def delete_resource(resource_id, version):
#     resource = database.get_resource_definition(resource_id, version)
#     resource.deleted = True
#     resource.active = False
#     db.session.update(resource)
#     db.session.commit()


# def set_permissions(resource_id: str, version: int, permissions: Dict[str, bool]):
#     resource_def = database.get_resource_definition(resource_id, version)
#     config = json.loads(resource_def.config_file)
#     config["protected"] = permissions
#     resource_def.config_file = json.dumps(config)
#     # db.session.update(resource_def)
#     db.session.commit()


# def is_protected(resource_id, level):
#     """
#     Level can be READ, WRITE or ADMIN
#     """
#     resource = get_resource(resource_id)
#     protection = resource.config.get("protected", {})
#     return level == "WRITE" or level == "ADMIN" or protection.get("read")


# def get_all_metadata(resource_obj: Resource) -> Dict[str, EntryMetadata]:
#     history_table = resource_obj.history_model
#     result = db.session.query(
#         history_table.entry_id,
#         history_table.user_id,
#         history_table.timestamp,
#         func.max(history_table.version),
#     ).group_by(history_table.entry_id)
#     result_ = {
#         row[0]: EntryMetadata(row[1], last_modified=row[2], version=row[3])
#         for row in result
#     }
#     return result_


# def get_metadata(resource_def: Resource, _id: int) -> EntryMetadata:
#     history_table = resource_def.history_model
#     result = (
#         db.session.query(
#             history_table.user_id,
#             history_table.timestamp,
#             func.max(history_table.version),
#         )
#         .filter(history_table.entry_id == _id)
#         .group_by(history_table.entry_id)
#     )
#     return EntryMetadata(result[0][0], last_modified=result[0][1], version=result[0][2])



import pytest

from karp.foundation.errors import ConsistencyError
from karp.foundation.entity import (
    TimestampedEntity,
    TimestampedVersionedEntity,
    VersionedEntity,
)


def test_versioned_entity_discard_w_wrong_version_raises_consistency_error():
    entity = VersionedEntity("id_v1", 1)

    # event = VersionedEntity.Discarded(entity_id=entity.id, entity_version=2)

    with pytest.raises(ConsistencyError):
        entity.discard(version=2)


def test_versioned_entity_stamped_w_wrong_version_raises_consistency_error():
    entity = VersionedEntity("id_v1", 1)

    event = VersionedEntity.Stamped(entity_id=entity.id, entity_version=2)

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_versioned_entity_stamped_increments_version():
    entity = VersionedEntity("id_v1", 1)

    event = VersionedEntity.Stamped(entity_id=entity.id, entity_version=entity.version)

    event.mutate(entity)

    assert entity.version == 2


def test_timestamped_entity_discard_w_wrong_last_modified_raises_consistency_error():
    entity = TimestampedEntity("id_v1", 1)

    event = TimestampedEntity.Discarded(
        entity_id=entity.id, entity_last_modified=2, user="Test"
    )

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_timestamped_entity_stamped_w_wrong_last_modified_raises_consistency_error():
    entity = TimestampedEntity("id_v1", 1)

    event = TimestampedEntity.Stamped(
        entity_id=entity.id, entity_last_modified=2, user="Test"
    )

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_timestamped_entity_stamped_updates_last_modified_and_last_modified_by():
    entity = TimestampedEntity("id_v1")

    event = TimestampedEntity.Stamped(
        entity_id=entity.id, entity_last_modified=entity.last_modified, user="Test"
    )

    previous_last_modified = entity.last_modified
    previous_last_modified_by = entity.last_modified_by
    event.mutate(entity)

    assert entity.last_modified > previous_last_modified
    assert entity.last_modified_by != previous_last_modified_by


def test_timestamped_versioned_entity():
    entity = TimestampedVersionedEntity("id1", version=1)

    assert isinstance(entity, TimestampedVersionedEntity)
    assert isinstance(entity, VersionedEntity)
    assert isinstance(entity, TimestampedEntity)


def test_timestamped_versioned_entity_discarded():
    entity = TimestampedVersionedEntity("id1", version=1)

    event = TimestampedVersionedEntity.Discarded(
        entity_id=entity.id,
        entity_version=entity.version,
        entity_last_modified=entity.last_modified,
        user="Test",
    )

    previous_last_modified = entity.last_modified
    previous_last_modified_by = entity.last_modified_by
    previous_version = entity.version
    event.mutate(entity)

    assert entity.discarded
    assert entity.last_modified > previous_last_modified
    assert entity.last_modified_by != previous_last_modified_by
    assert entity.version == (previous_version + 1)


def test_timestamped_versioned_entity_discarded_w_wrong_version_raises_consistency_error():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    event = TimestampedVersionedEntity.Discarded(entity_id=entity.id, entity_version=2)

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_timestamped_versioned_entity_discarded_w_wrong_last_modified_raises_consistency_error():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    event = TimestampedVersionedEntity.Discarded(
        entity_id=entity.id, entity_version=1, entity_last_modified=2, user="Test"
    )

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_timestamped_versioned_entity_stamped_w_wrong_version_raises_consistency_error():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    event = TimestampedVersionedEntity.Stamped(entity_id=entity.id, entity_version=2)

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_timestamped_versioned_entity_stamped_w_wrong_last_modified_raises_consistency_error():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    event = TimestampedVersionedEntity.Stamped(
        entity_id=entity.id, entity_version=1, entity_last_modified=2, user="Test"
    )

    with pytest.raises(ConsistencyError):
        event.mutate(entity)


def test_timestamped_versioned_entity_stamped_updates():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    event = TimestampedVersionedEntity.Stamped(
        entity_id=entity.id,
        entity_version=entity.version,
        entity_last_modified=entity.last_modified,
        user="Test",
    )

    previous_last_modified = entity.last_modified
    previous_last_modified_by = entity.last_modified_by
    previous_version = entity.version

    event.mutate(entity)

    assert entity.last_modified > previous_last_modified
    assert entity.last_modified_by != previous_last_modified_by
    assert entity.version == (previous_version + 1)


def test_timestamped_versioned_entity_stamp_updates_last_modified_and_version():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    previous_last_modified = entity.last_modified
    previous_last_modified_by = entity.last_modified_by
    previous_version = entity.version

    entity.stamp(user="Test")

    assert entity.last_modified > previous_last_modified
    assert entity.last_modified_by == "Test"
    assert entity.version == (previous_version + 1)


def test_timestamped_versioned_entity_stamp_w_increment_version_eq_false_updates_last_modified_and_not_version():
    entity = TimestampedVersionedEntity("id_v1", version=1)

    previous_last_modified = entity.last_modified
    previous_last_modified_by = entity.last_modified_by
    previous_version = entity.version

    entity.stamp(user="Test", increment_version=False)

    assert entity.last_modified > previous_last_modified
    assert entity.last_modified_by == "Test"
    assert entity.version == previous_version



"""Constraints."""


class ConstraintsError(ValueError):
    """Raised when a constraint is not met."""

    pass


def length_gt_zero(attribute, value):
    if len(value) == 0:
        raise ConstraintsError(
            f"'{attribute}' has to be non-empty. Got {attribute}='{value}'"
        )
    return value


def valid_resource_id(name: str):
    return length_gt_zero("resource_id", name)



"""Handle of unique ids.

Borrowed from https://bitbucket.org/sixty-north/d5-kanban-python
"""
import uuid
import ulid
import typing
try:
    import fastuuid  # type: ignore
except ModuleNotFoundError:
    fastuuid = None


def _make_unique_id() -> uuid.UUID:
    """Make a new UniqueId."""
    return ulid.new().uuid
    # return UniqueId(uuid.uuid4())


if fastuuid is not None:
    make_unique_id = fastuuid.uuid4
    UniqueId = fastuuid.UUID
    UniqueIdType = (fastuuid.UUID, uuid.UUID)
    typing_UniqueId = typing.Union[fastuuid.UUID, uuid.UUID]
else:
    make_unique_id = _make_unique_id
    UniqueId = uuid.UUID
    UniqueIdType = uuid.UUID
    typing_UniqueId = uuid.UUID
# class UniqueId:
#     """A UUID-based unique id with global formatting control.

#     When used in format specifiers the number of leading characters
#     of the UUID can be specified:

#         message = "The id={:6}".format(my_unique_id)

#     If not specified the globally configured format length will be
#     used. The global format length can be set by adjusting the
#     abbreviated_length class attribute. Setting this value to None
#     (the default) disables global format abbreviation.
#     """

#     abbreviated_length = None

#     @classmethod
#     def from_hex(cls, hex_str):
#         """Construct a UniqueId from a hex string representation of a UUID."""
#         return cls(uuid.UUID(hex=hex_str))

#     def __init__(self, the_uuid):
#         """Instantiate a UniqueId using a existing UUID object.

#         To create UniqueIds with fresh UUIDs call the make_unique_id()
#         factory function instead.

#         Args:
#             the_uuid: A UUID object.
#         """
#         self._uuid = the_uuid

#     def __eq__(self, rhs):
#         if not isinstance(rhs, UniqueId):
#             return NotImplemented
#         return self._uuid == rhs._uuid

#     def __ne__(self, rhs):
#         if not isinstance(rhs, UniqueId):
#             return NotImplemented
#         return self._uuid != rhs._uuid

#     def __hash__(self):
#         return hash(self._uuid)

#     def __repr__(self):
#         return self._uuid.hex

#     def __format__(self, format_spec):
#         if not format_spec:
#             return repr(self)[: UniqueId.abbreviated_length]
#         try:
#             format_abbrev_length = int(format_spec)
#         except ValueError:
#             raise TypeError(
#                 "UniqueId format spec {!r} is not an integer length".format(format_spec)
#             )
#         return repr(self)[:format_abbrev_length]



import pytest

from karp.lex_infrastructure.repositories.sql_entries import SqlEntryRepository

from karp.tests.unit.lex import factories


@pytest.fixture(name="entry_repo")
def fixture_entry_repo(sqlite_session_factory):
    session = sqlite_session_factory()
    entry_repo = SqlEntryRepository.from_dict(
        name="test_name",
        resource_config={},
        session=session,
    )
    yield entry_repo

    entry_repo.teardown()


@pytest.fixture(name="entry_repo2")
def fixture_entry_repo2(sqlite_session_factory):
    session = sqlite_session_factory()
    entry_repo = SqlEntryRepository.from_dict(
        name="test_name2",
        resource_config={},
        session=session,
    )
    yield entry_repo

    entry_repo.teardown()


def test_create_entry_repository(entry_repo):
    assert entry_repo.entry_ids() == []


def test_save_entry_to_entry_repo(entry_repo):
    entry = factories.EntryFactory()
    entry_repo.save(entry)

    assert entry_repo.by_id(entry.id).entry_id == entry.entry_id
    assert entry_repo.by_entry_id(entry.entry_id).id == entry.id


def test_update_entry_to_entry_repo(entry_repo):
    entry = factories.EntryFactory()
    entry_repo.save(entry)

    entry.stamp(user='kristoff@example.com', message='hi')
    entry_repo.save(entry)

    assert entry_repo.by_id(entry.id).entry_id == entry.entry_id
    assert entry_repo.by_id(entry.id).version == entry.version
    assert entry_repo.by_entry_id(entry.entry_id).id == entry.id
    assert entry_repo.by_entry_id(entry.entry_id).version == entry.version


def test_create_entry_repository2(entry_repo2):
    assert entry_repo2.entry_ids() == []



import collections
import json
import logging
import typing
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Generic, List, Optional, Tuple

import fastjsonschema  # pyre-ignore
import json_streams
from sb_json_tools import jsondiff
import logging

from karp.foundation.commands import CommandHandler
from karp.lex.application.repositories import EntryUnitOfWork
from karp.lex.domain import errors, entities
from karp.lex.domain.entities.entry import Entry
from karp.lex.domain.entities.resource import Resource
from karp.foundation.value_objects import unique_id
from karp.errors import (ClientErrorCodes, EntryIdMismatch, EntryNotFoundError,
                         KarpError, UpdateConflict)
from karp.foundation import events as foundation_events
from karp.foundation import messagebus
from karp.lex.domain import commands
from karp.lex.domain.value_objects import EntrySchema
from karp.lex.application import repositories


logger = logging.getLogger(__name__)


# def get_entries_by_column(resource_obj: Resource, filters):
#     config = resource_obj.config
#     cls = resource_obj.model
#     query = cls.query
#
#     joined_filters = []
#     simple_filters = {}
#
#     for filter_key in filters.keys():
#         tmp = collections.defaultdict(dict)
#         if filter_key in config["referenceable"] and config["fields"][filter_key].get(
#             "collection", False
#         ):
#             child_cls = cls.child_tables[filter_key]
#             tmp[child_cls.__tablename__][filter_key] = filters[filter_key]
#         else:
#             simple_filters[filter_key] = filters[filter_key]
#         joined_filters.extend(tmp.values())
#
#     query = query.filter_by(**simple_filters)
#
#     for child_filters in joined_filters:
#         child_cls = cls.child_tables[list(child_filters.keys())[0]]
#         query = query.join(child_cls).filter_by(**child_filters)
#
#     return [
#         {
#             "id": db_entry.id,
#             "entry_id": db_entry.entry_id,
#             "entry": json.loads(db_entry.body),
#         }
#         for db_entry in query
#     ]
#
#
# def get_entry(resource_id: str, entry_id: str, version: Optional[int] = None):
#     resource = get_resource(resource_id, version=version)
#     return get_entry_by_entry_id(resource, entry_id)
#
#
# def get_entry_by_entry_id(resource: Resource, entry_id: str):
#     cls = resource.model
#     return cls.query.filter_by(entry_id=entry_id).first()
#
#
class BasingEntry:
    def __init__(
        self,
        entry_repo_uow: repositories.EntryUowRepositoryUnitOfWork,
        resource_uow: repositories.ResourceUnitOfWork
    ) -> None:
        super().__init__()
        self.entry_repo_uow = entry_repo_uow
        self.resource_uow = resource_uow

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()
        yield from self.entry_repo_uow.collect_new_events()

    def get_entry_uow(self, entry_repo_id: unique_id.UniqueId) -> EntryUnitOfWork:
        with self.entry_repo_uow as uw:
            return uw.repo.get_by_id(entry_repo_id)


class AddingEntry(BasingEntry, CommandHandler[commands.AddEntry]):

    def execute(self, cmd: commands.AddEntry):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(
                cmd.resource_id)

        try:
            entry_id = resource.id_getter()(cmd.entry)
        except KeyError as err:
            raise errors.MissingIdField(
                resource_id=cmd.resource_id,
                entry=cmd.entry
            ) from err
        entry_schema = EntrySchema(resource.entry_json_schema)

        with self.get_entry_uow(resource.entry_repository_id) as uw:
            existing_entry = uw.repo.get_by_entry_id_optional(entry_id)
            if (
                existing_entry
                and not existing_entry.discarded
                and existing_entry.id != cmd.entity_id
            ):
                raise errors.IntegrityError(
                    f"An entry with entry_id '{entry_id}' already exists."
                )

            entry_schema.validate_entry(cmd.entry)

            entry = resource.create_entry_from_dict(
                cmd.entry,
                user=cmd.user,
                message=cmd.message,
                entity_id=cmd.entity_id
            )
            uw.entries.save(entry)
            uw.commit()
        return entry


#
# def preview_entry(resource_id, entry, resource_version=None):
#     resource = get_resource(resource_id, version=resource_version)
#     entry_json = _validate_and_prepare_entry(resource, entry)
#     return entry_json


class UpdatingEntry(BasingEntry, CommandHandler[commands.UpdateEntry]):
    def execute(self, cmd: commands.UpdateEntry):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(
                cmd.resource_id
            )

        entry_schema = EntrySchema(resource.entry_json_schema)
        entry_schema.validate_entry(cmd.entry)

        with self.get_entry_uow(resource.entry_repository_id) as uw:
            try:
                current_db_entry = uw.repo.by_entry_id(
                    cmd.entry_id
                )
            except errors.EntryNotFound as err:
                raise errors.EntryNotFound(
                    cmd.resource_id,
                    cmd.entry_id,
                    entity_id=None,
                ) from err

            diff = jsondiff.compare(current_db_entry.body, cmd.entry)
            if not diff:
                return current_db_entry

            #     db_entry_json = json.dumps(entry)
            #     db_id = current_db_entry.id
            #     latest_history_entry = (
            #         resource.history_model.query.filter_by(entry_id=db_id)
            #         .order_by(resource.history_model.version.desc())
            #         .first()
            #     )
            if not cmd.force and current_db_entry.version != cmd.version:
                logger.info(
                    'version conflict', current_version=current_db_entry.version, version=cmd.version)
                raise errors.UpdateConflict(diff)

            id_getter = resource.id_getter()
            new_entry_id = id_getter(cmd.entry)

            current_db_entry.body = cmd.entry
            current_db_entry.stamp(
                cmd.user, message=cmd.message, timestamp=cmd.timestamp)
            if new_entry_id != cmd.entry_id:
                logger.info('updating entry_id',
                            entry_id=cmd.entry_id, new_entry_id=new_entry_id)
                current_db_entry.entry_id = new_entry_id
                # uw.repo.move(current_db_entry, old_entry_id=cmd.entry_id)
                uw.repo.save(current_db_entry)
            else:
                uw.repo.save(current_db_entry)
            uw.commit()

        return current_db_entry


def update_entries(*args, **kwargs):
    return []


def add_entries_from_file(
    resource_id: str, version: int, filename: Path
) -> List[Entry]:
    return add_entries(
        resource_id,
        json_streams.load_from_file(filename),
        user_id="local_admin",
        resource_version=version,
    )


class AddingEntries(
    BasingEntry,
    CommandHandler[commands.AddEntries]
):
    def execute(self, cmd: commands.AddEntries):
        """
        Add entries to DB and INDEX (if present and resource is active).

        Raises
        ------
        RuntimeError
            If the resource.entry_json_schema fails to compile.
        KarpError
            - If an entry fails to be validated against the json schema.
            - If the DB interaction fails.

        Returns
        -------
        List
            List of the id's of the created entries.
        """

        if not isinstance(cmd.resource_id, str):
            raise ValueError(
                f"'resource_id' must be of type 'str', were '{type(cmd.resource_id)}'"
            )
        with self.resource_uow:
            resource = self.resource_uow.resources.by_resource_id(
                cmd.resource_id)

        if not resource:
            raise errors.ResourceNotFound(cmd.resource_id)

        entry_schema = EntrySchema(resource.entry_json_schema)

        created_db_entries = []
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
                resource.entry_repository_id) as uw:
            for entry_raw in cmd.entries:
                entry_schema.validate_entry(entry_raw)

                entry = resource.create_entry_from_dict(
                    entry_raw,
                    user=cmd.user,
                    message=cmd.message,
                    entity_id=unique_id.make_unique_id(),
                )
                uw.entries.save(entry)
                created_db_entries.append(entry)
            uw.commit()

        return created_db_entries


# def add_entries(
#     resource_id: str,
#     entries: List[Dict],
#     user_id: str,
#     message: str = None,
#     resource_version: int = None,
# ):
#     """
#     Add entries to DB and INDEX (if present and resource is active).
#
#     Raises
#     ------
#     RuntimeError
#         If the resource.entry_json_schema fails to compile.
#     KarpError
#         - If an entry fails to be validated against the json schema.
#         - If the DB interaction fails.
#
#     Returns
#     -------
#     List
#         List of the id's of the created entries.
#     """
#     resource = get_resource(resource_id, version=resource_version)
#     resource_conf = resource.config
#
#     validate_entry = _compile_schema(resource.entry_json_schema)
#
#     try:
#         created_db_entries = []
#         for entry in entries:
#             _validate_entry(validate_entry, entry)
#
#             entry_json = json.dumps(entry)
#             db_entry = _src_entry_to_db_entry(
#                 entry, entry_json, resource.model, resource_conf
#             )
#             created_db_entries.append((db_entry, entry, entry_json))
#             db.session.add(db_entry)
#
#         db.session.commit()
#
#         created_history_entries = []
#         for db_entry, entry, entry_json in created_db_entries:
#             history_entry = resource.history_model(
#                 entry_id=db_entry.id,
#                 user_id=user_id,
#                 body=entry_json,
#                 version=1,
#                 op="ADD",
#                 message=message,
#                 timestamp=datetime.now(timezone.utc).timestamp(),
#             )
#             created_history_entries.append((db_entry, entry, history_entry))
#             db.session.add(history_entry)
#         db.session.commit()
#     except sql_exception.IntegrityError as e:
#         _logger.exception("IntegrityError")
#         print("e = {e!r}".format(e=e))
#         print("e.orig.args = {e!r}".format(e=e.orig.args))
#         raise KarpError(
#             "Database error: {msg}".format(msg=e.orig.args),
#             ClientErrorCodes.DB_INTEGRITY_ERROR,
#         )
#     except sql_exception.SQLAlchemyError as e:
#         _logger.exception("Adding entries to DB failed.")
#         print("e = {e!r}".format(e=e))
#         raise KarpError(
#             "Database error: {msg}".format(msg=e.msg), ClientErrorCodes.DB_GENERAL_ERROR
#         )
#
#     if resource.active:
#         indexmgr.add_entries(
#             resource_id,
#             [
#                 (
#                     db_entry.entry_id,
#                     entrymetadata.EntryMetadata.init_from_model(history_entry),
#                     _src_entry_to_index_entry(resource, entry),
#                 )
#                 for db_entry, entry, history_entry in created_history_entries
#             ],
#         )
#
#     return [db_entry.entry_id for db_entry, _, _ in created_db_entries]
#
#
# def _src_entry_to_db_entry(entry, entry_json, resource_model, resource_conf):
#     kwargs = _src_entry_to_db_kwargs(entry, entry_json, resource_model, resource_conf)
#     db_entry = resource_model(**kwargs)
#     return db_entry
#
#
# def _src_entry_to_db_kwargs(entry, entry_json, resource_model, resource_conf):
#     kwargs = {"body": entry_json}
#
#     for field_name in resource_conf.get("referenceable", ()):
#         field_val = entry.get(field_name)
#         if resource_conf["fields"][field_name].get("collection", False):
#             child_table = resource_model.child_tables[field_name]
#             for elem in field_val:
#                 if field_name not in kwargs:
#                     kwargs[field_name] = []
#                 kwargs[field_name].append(child_table(**{field_name: elem}))
#         else:
#             if field_val:
#                 kwargs[field_name] = field_val
#     id_field = resource_conf.get("id")
#     if id_field:
#         kwargs["entry_id"] = entry[id_field]
#     else:
#         kwargs["entry_id"] = "TODO"  # generate id for resources that are missing it
#     return kwargs


class DeletingEntry(BasingEntry, CommandHandler[commands.DeleteEntry]):

    def execute(self, cmd: commands.DeleteEntry):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(cmd.resource_id)

        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            resource.entry_repository_id
        ) as uw:
            entry = uw.repo.by_entry_id(cmd.entry_id)

            entry.discard(
                user=cmd.user,
                message=cmd.message,
                timestamp=cmd.timestamp,
            )
            uw.repo.save(entry)
            uw.commit()

    # ctx.search_service.delete_entry(resource, entry=entry)


# def delete_entry(resource_id: str, entry_id: str, user_id: str):
#     resource = get_resource(resource_id)
#     entry = resource.model.query.filter_by(entry_id=entry_id, deleted=False).first()
#     if not entry:
#         raise EntryNotFoundError(resource_id, entry_id)
#     entry.deleted = True
#     history_cls = resource.history_model
#     history_entry = history_cls(
#         entry_id=entry.id,
#         user_id=user_id,
#         op="DELETE",
#         version=-1,
#         timestamp=datetime.now(timezone.utc).timestamp(),
#     )
#     db.session.add(history_entry)
#     db.session.commit()
#     indexmgr.delete_entry(resource_id, entry.entry_id)
#
#
def _src_entry_to_index_entry(resource: Resource, src_entry: Entry) -> Dict:
    return indexing.transform_to_index_entry(
        ctx.resource_repo, ctx.search_service, resource, src_entry
    )


# def _src_entry_to_index_entry(resource: Resource, src_entry: Dict):
#     """
#     Make a "src entry" into an "index entry"
#     """
#     return indexmgr.transform_to_index_entry(
#         resource, src_entry, resource.config["fields"].items()
#     )
#
#
# def _validate_and_prepare_entry(resource, entry):
#     validate_entry = _compile_schema(resource.entry_json_schema)
#     _validate_entry(validate_entry, entry)
#     return _src_entry_to_index_entry(resource, entry)



# import json  # noqa: F401

from karp.utility.json_schema import create_entry_json_schema

CONFIG_PLACES = """{
  "resource_id": "places",
  "resource_name": "Platser i Sverige",
  "fields": {
    "name": {
      "type": "string",
      "required": true
    },
    "municipality": {
      "collection": true,
      "type": "number",
      "required": true
    },
    "population": {
      "type": "number"
    },
    "area": {
      "type": "number"
    },
    "density": {
      "type": "number"
    },
    "code": {
      "type": "number",
      "required": true
    }
  },
  "sort": "name",
  "id": "code"
}"""


def test_create_json_schema(json_schema_config):
    json_schema = create_entry_json_schema(json_schema_config["fields"])
    assert json_schema["type"] == "object"



"""Factories used in tests."""
import typing

import factory
from faker.providers import BaseProvider

from karp.lex.domain import commands, entities, events
from karp.foundation.value_objects import make_unique_id
from karp.utility.time import utc_now


class ResourceConfigProvider(BaseProvider):
    def resource_config(self) -> typing.Dict:
        return {
            "sort": ["baseform"],
            "fields": {
                "baseform": {"type": "string", "required": True},
                'wordclass': {'type': 'string'},
            },
            "id": "baseform",
        }


factory.Faker.add_provider(ResourceConfigProvider)


class ResourceFactory(factory.Factory):
    class Meta:
        model = entities.Resource

    entity_id = factory.LazyFunction(make_unique_id)
    entry_repo_id = factory.LazyFunction(make_unique_id)
    resource_id = factory.Faker("word")
    name = factory.Faker('word')
    config = factory.Faker('resource_config')
    last_modified_by = factory.Faker("email")
    last_modified = factory.LazyFunction(utc_now)
    message = 'resource created'


class EntryFactory(factory.Factory):
    class Meta:
        model = entities.Entry

    entity_id = factory.LazyFunction(make_unique_id)
    entry_id = factory.Faker("word")
    last_modified_by = factory.Faker("email")
    last_modified = factory.LazyFunction(utc_now)
    message = 'entry created'
    body = {}
    repository_id = factory.LazyFunction(make_unique_id)


def random_resource(config: typing.Optional[typing.Dict] = None):
    config = config or {"fields": {"wf": {"type" "string"}, "id": "wf"}}
    return entities.create_resource(
        entity_id=make_unique_id(),
        entry_repo_id=make_unique_id(),
        resource_id="resource",
        config=config,
        message="Resource add",
        created_by="kristoff@example.com",
    )


class ResourceCreatedFactory(factory.Factory):
    class Meta:
        model = events.ResourceCreated

    timestamp = factory.LazyFunction(utc_now)
    entity_id = factory.LazyFunction(make_unique_id)
    entry_repo_id = factory.LazyFunction(make_unique_id)
    resource_id = factory.Faker('word')
    name = factory.Faker('word')
    config = factory.Faker('resource_config')
    user = factory.Faker('email')
    message = 'resource created'


class ResourcePublishedFactory(factory.Factory):
    class Meta:
        model = events.ResourcePublished

    timestamp = factory.LazyFunction(utc_now)
    entity_id = factory.LazyFunction(make_unique_id)
    entry_repo_id = factory.LazyFunction(make_unique_id)
    resource_id = factory.Faker('word')
    name = factory.Faker('word')
    config = factory.Faker('resource_config')
    user = factory.Faker('email')
    version = factory.Sequence(int)
    message = 'resource created'


class EntryAddedFactory(factory.Factory):
    class Meta:
        model = events.EntryAdded

    timestamp = factory.LazyFunction(utc_now)
    id = factory.LazyFunction(make_unique_id)
    resource_id = factory.Faker('word')
    entry_id = factory.Faker('word')
    body = factory.Faker('resource_config')
    user = factory.Faker('email')
    message = 'resource created'


class CreateEntryRepositoryFactory(factory.Factory):
    class Meta:
        model = commands.CreateEntryRepository

    entity_id = factory.LazyFunction(make_unique_id)
    name = factory.Faker('word')
    repository_type = 'fake'
    config = factory.Faker('resource_config')
    message = 'entry repository created'
    user = factory.Faker('email')


class CreateResourceFactory(factory.Factory):
    class Meta:
        model = commands.CreateResource

    entity_id = factory.LazyFunction(make_unique_id)
    resource_id = factory.Faker('word')
    name = factory.Faker('word')
    repository_type = 'fake'
    config = factory.Faker('resource_config')
    entry_repo_id = factory.LazyFunction(make_unique_id)
    message = 'created'
    user = factory.Faker('email')


class UpdateResourceFactory(factory.Factory):
    class Meta:
        model = commands.UpdateResource

    resource_id = factory.Faker('word')
    name = factory.Faker('word')
    repository_type = 'fake'
    config = factory.Faker('resource_config')
    entry_repo_id = factory.LazyFunction(make_unique_id)
    message = 'created'
    created_by = 'kristoff@example.com'


class AddEntryFactory(factory.Factory):
    class Meta:
        model = commands.AddEntry

    entity_id = factory.LazyFunction(make_unique_id)
    resource_id = factory.Faker('word')
    entry = {
        'baseform': 'bra',
    }
    user = factory.Faker('email')
    message = 'added'


class UpdateEntryFactory(factory.Factory):
    class Meta:
        model = commands.UpdateEntry

    resource_id = factory.Faker('word')
    entry_id = factory.Faker('word')
    entry = {}
    user = factory.Faker('email')
    message = 'added'


class DeleteEntryFactory(factory.Factory):
    class Meta:
        model = commands.DeleteEntry

    resource_id = factory.Faker('word')
    entry_id = factory.Faker('word')
    user = factory.Faker('email')
    message = 'deleted'
